<?php

namespace Drupal\performance_profiler\Form;

use Drupal\Core\Database\Connection;
use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\performance_profiler\PerformanceDatabaseActions;
use Psr\Log\LoggerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Configuration form for Database benchmark tool setting.
 */
class PerformanceProfilerDbForm extends FormBase {

  /**
   * The database connection.
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected $database;

  /**
   * Performance profiler Database actions service.
   *
   * @var \Drupal\performance_profiler\PerformanceDatabaseActions
   */
  protected $dbActions;

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'performance_profiler_database';
  }

  /**
   * Class construct.
   *
   * @param \Drupal\Core\Database\Connection $database
   *   State interface service.
   * @param \Drupal\performance_profiler\PerformanceDatabaseActions $db_actions
   *   Performance profiler Database actions service.
   */
  public function __construct(Connection $database, PerformanceDatabaseActions $db_actions) {
    $this->database = $database;
    $this->dbActions = $db_actions;
  }

  /**
   * Factory method for dependency injection container.
   *
   * @param \Symfony\Component\DependencyInjection\ContainerInterface $container
   *   Container.
   *
   * @return static
   *   Return static.
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('database'),
      $container->get('performance_profiler.database_actions'),
    );
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state): array {
    // Make sure we do not accidentally cache this form.
    $form['#cache']['max-age'] = 0;

    $form['benchmarks'] = [
      '#type' => 'container',
    ];

    $form['benchmarks']['action'] = [
      '#type' => 'select',
      '#title' => $this->t('Action'),
      '#required' => TRUE,
      '#options' => [
        'create' => $this->t('Create table'),
        'fill' => $this->t('Fill table'),
        'query' => $this->t('Run query'),
        'truncate' => $this->t('Truncate table'),
        'drop' => $this->t('Drop table'),
      ],
      '#empty_option' => $this->t('- Select action -'),
    ];

    $tables = PerformanceDatabaseActions::tables();

    $tables = array_keys($tables);
    $tables = array_combine($tables, $tables);
    $tables_to_operate = $tables;
    foreach ($tables_to_operate as $table => $label) {
      $table_name = PerformanceDatabaseActions::TABLE_PREFIX . $table;
      $tables_to_operate[$table] = $table_name;
      $tables[$table] = $table_name;
      if (!$this->database->schema()->tableExists($table_name)) {
        unset($tables_to_operate[$table]);
      }
      else {
        unset($tables[$table]);
      }
    }

    $form['benchmarks']['tables_create'] = [
      '#type' => 'item',
      '#markup' => $this->t('All tables already created.'),
      '#states' => [
        'visible' => [
          [':input[name="action"]' => ['value' => 'create']],
        ],
      ],
    ];
    if (!empty($tables)) {
      unset($form['benchmarks']['tables_create']['#markup']);
      $form['benchmarks']['tables_create']['#type'] = 'checkboxes';
      $form['benchmarks']['tables_create']['#title'] = $this->t('Tables');
      $form['benchmarks']['tables_create']['#options'] = $tables;
    }

    $form['benchmarks']['tables'] = [
      '#type' => 'item',
      '#markup' => $this->t('There are no created tables, please run create action first.'),
      '#states' => [
        'visible' => [
          [':input[name="action"]' => ['value' => 'fill']],
          [':input[name="action"]' => ['value' => 'truncate']],
          [':input[name="action"]' => ['value' => 'drop']],
        ],
      ],
    ];
    if (!empty($tables_to_operate)) {
      unset($form['benchmarks']['tables']['#markup']);
      $form['benchmarks']['tables']['#type'] = 'checkboxes';
      $form['benchmarks']['tables']['#title'] = $this->t('Tables');
      $form['benchmarks']['tables']['#options'] = $tables_to_operate;

      $form['benchmarks']['cycle_insert'] = [
        '#type' => 'radios',
        '#title' => $this->t('Insert type'),
        '#options' => [
          'one_operation' => $this->t('Insert values by one operation'),
          'cycle' => $this->t('Insert values in cycle one by one'),
          'transaction' => $this->t('Insert values in cycle one by one using transaction'),
        ],
        '#default_value' => 'one_operation',
        '#states' => [
          'visible' => [
            [':input[name="action"]' => ['value' => 'fill']],
          ],
        ],
      ];
    }

    $form['benchmarks']['query'] = [
      '#type' => 'select',
      '#title' => $this->t('Query'),
      '#options' => [
        'all' => $this->t('Run all'),
        'Main table operations' => [
          'select' => $this->t('Select all items'),
          'select_sort' => $this->t('Select 1000 items, sorted by indexed column'),
          'select_sort_no_index' => $this->t('Select 1000 items, sorted by non-indexed column'),
          'select_like' => $this->t('Select 1000 items using LIKE "%ab%" condition'),
          'select_group_by' => $this->t('Select 1000 items with group by'),
        ],
        'Several tables operations' => [
          'join' => $this->t('Select all items using join by uuid'),
          'join_two' => $this->t('Select all items using two joined tables'),
        ],
      ],
      '#states' => [
        'visible' => [
          [':input[name="action"]' => ['value' => 'query']],
        ],
      ],
    ];

    $form['actions']['#type'] = 'actions';
    $form['actions']['execute'] = [
      '#type' => 'submit',
      '#value' => $this->t('Execute'),
      '#button_type' => 'primary',
      '#submit' => ['::executeAction'],
    ];

    return $form;
  }

  /**
   * Profile create mySQL tables action.
   */
  public function executeAction(array &$form, FormStateInterface $form_state) {
    $action = $form_state->getValue('action');
    switch ($action) {
      case 'create':
        $tables = $form_state->getValue('tables_create');
        if (!empty($tables) && is_array($tables)) {
          $tables = array_filter($tables);
        }
        if (!empty($tables)) {
          $specifications = PerformanceDatabaseActions::tables();
          foreach ($tables as $table => $value) {
            $this->dbActions->createTable($table, $specifications[$table]);
          }
        }
        break;

      case 'truncate':
      case 'drop':
      case 'fill':
        $tables = $form_state->getValue('tables');
        if (!empty($tables) && is_array($tables)) {
          $tables = array_filter($tables);
        }
        if (!empty($tables)) {
          if ($action == 'fill') {
            $one_operation = $form_state->getValue('cycle_insert');
          }
          $callback = $action . 'Table';
          foreach ($tables as $table => $value) {
            $this->dbActions->{$callback}($table, $one_operation ?? NULL);
          }
        }
        break;

      case 'query':
        $query_type = $form_state->getValue('query');
        if ($query_type == 'all') {
          foreach (PerformanceDatabaseActions::QUERY_TYPES as $query_type) {
            $this->dbActions->runQuery($query_type);
          }
        }
        else {
          $this->dbActions->runQuery($query_type);
        }
        break;
    }

    $this->dbActions->memoryUsage();
  }

  /**
   * {@inheritDoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
  }

}
