<?php

namespace Drupal\performance_profiler\Form;

use Drupal\Core\Database\Connection;
use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\performance_profiler\PerformanceBenchmark;
use Drupal\performance_profiler\PerformanceDatabaseActions;
use Psr\Log\LoggerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Configure Memory profiler module settings.
 *
 * @internal
 */
class PerformanceProfilerRunForm extends FormBase {

  /**
   * The database connection.
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected $database;

  /**
   * Messenger instance.
   *
   * @var \Drupal\Core\Messenger\MessengerInterface
   */
  protected $messenger;

  /**
   * Performance profiler Database actions service.
   *
   * @var \Drupal\performance_profiler\PerformanceDatabaseActions
   */
  protected $dbActions;

  /**
   * Class construct.
   *
   * @param \Drupal\Core\Database\Connection $database
   *   State interface service.
   * @param \Drupal\Core\Messenger\MessengerInterface $messenger
   *   Messenger service.
   * @param \Drupal\performance_profiler\PerformanceDatabaseActions $db_actions
   *   Performance profiler Database actions service.
   */
  public function __construct(Connection $database, MessengerInterface $messenger, PerformanceDatabaseActions $db_actions) {
    $this->database = $database;
    $this->messenger = $messenger;
    $this->dbActions = $db_actions;
  }

  /**
   * Factory method for dependency injection container.
   *
   * @param \Symfony\Component\DependencyInjection\ContainerInterface $container
   *   Container.
   *
   * @return static
   *   Return static.
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('database'),
      $container->get('messenger'),
      $container->get('performance_profiler.database_actions'),
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'performance_profiler_run';
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    $form['actions']['#type'] = 'actions';

    $form['actions']['submit'] = [
      '#type' => 'submit',
      '#value' => $this->t('Run benchmarks'),
      '#button_type' => 'primary',
    ];

    $form['actions']['bench_php'] = [
      '#type' => 'submit',
      '#value' => $this->t('Run PHP benchmark'),
      '#button_type' => 'secondary',
      '#submit' => ['::benchmarkPhp'],
    ];

    $form['actions']['bench_database'] = [
      '#type' => 'submit',
      '#value' => $this->t('Run Database benchmark'),
      '#button_type' => 'secondary',
      '#submit' => ['::benchmarkDb'],
    ];

    return $form;
  }

  /**
   * Submit handler for PHP benchmark.
   */
  public function benchmarkPhp(array &$form, FormStateInterface $form_state) {
    foreach (PerformanceBenchmark::run(FALSE) as $value) {
      \Drupal::messenger()->addMessage($value);
    }
  }

  /**
   * Submit handler for Database benchmark.
   */
  public function benchmarkDb(array &$form, FormStateInterface $form_state) {
    try {
      $tables = PerformanceDatabaseActions::tables();
      $table_ids = array_keys($tables);

      // Drop tables if exists.
      $phase = 'initial drop';
      foreach ($table_ids as $table) {
        $table_name = PerformanceDatabaseActions::TABLE_PREFIX . $table;
        if ($this->database->schema()->tableExists($table_name)) {
          $this->dbActions->dropTable($table);
        }
      }

      $phase = 'create tables';
      // Create tables if not exists.
      foreach ($tables as $table => $specification) {
        $this->dbActions->createTable($table, $specification);
      }

      $phase = 'fill tables';
      // Fill tables if not exists.
      $table_ids = array_keys($tables);
      foreach ($table_ids as $table) {
        $this->dbActions->fillTable($table);
      }

      $phase = 'run queries';
      // Run queries.
      foreach (PerformanceDatabaseActions::QUERY_TYPES as $query_type) {
        $this->dbActions->runQuery($query_type);
      }

      $phase = 'drop tables';
      // Drop tables at the end.
      $table_ids = array_keys($tables);
      foreach ($table_ids as $table) {
        $this->dbActions->dropTable($table);
      }

      $this->dbActions->memoryUsage();
    }
    catch (\Exception $e) {
      $this->messenger->addWarning("Failed executing query (phase: $phase), " . $e->getMessage());
    }
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $this->benchmarkPhp($form, $form_state);
    $this->benchmarkDb($form, $form_state);
  }

}
